/**
 *
 * @authors yutent (yutent@doui.cc)
 * @date    2017-04-14 21:04:50
 *
 */
import { bind, fire } from 'wkit'
import { noop, query2object, object2query } from '../utils.js'
import { __ROUTER_VIEW__ } from '../constants.js'

//hash前缀正则
const PREFIX_REGEXP = /^(#!|#)[\/]+?/
const RULE_REGEXP = /(\/[^/]*)(:[A-Za-z0-9_]+)(\?)?/g

const MODE_HASH = 'hash'
const MODE_HISTORY = 'history'

class Router {
  type = MODE_HASH

  #tables = new Map()
  #views = new Set()

  #targets = new Map()

  #ready = false
  #route = Object.create(null)

  #beforeEach

  constructor(type = MODE_HASH) {
    this.type = type
    bind(window, 'popstate', this.#hashchange.bind(this))
  }

  // 只读route
  get route() {
    return this.#route
  }

  // 只读views
  get views() {
    return Array.from(this.#views)
  }

  #hashchange(ev) {
    if (!this.#ready) {
      return
    }
    this.#check()
  }

  #parseRule(route) {
    if (route.path === '!') {
      route.regexp = null
    } else {
      let vars = []
      let re
      if (route.path.includes('?') && route.path.at(-1) !== '?') {
        throw new SyntaxError(
          `The exp "?" can only be used in the last.\n\n  ${JSON.stringify(
            route
          )}\n`
        )
      }

      re = route.path.replace(
        RULE_REGEXP,
        function (m, _prefix, _var, _require) {
          vars.push(_var.slice(1))
          if (_prefix === '/') {
            _prefix = '/?'
          }
          return _prefix + '([A-Za-z0-9_]+)' + _require
        }
      )

      re = '^' + re + '$'
      route.regexp = new RegExp(re)
      route.vars = vars
    }
    return route
  }

  #add(route) {
    if (route.path !== '!' && route.path[0] !== '/') {
      console.error('route path must start with "/"')
      return
    }
    route.path = route.path.replace(/^[\/]+|[\/]+$|\s+/g, '/')

    this.#tables.set(route.path, this.#parseRule(route))
    this.#views.add(route.name)
  }

  // 路由检测
  #check() {
    let isHash = this.type === MODE_HASH
    let $view = window.wkitd.get(__ROUTER_VIEW__)
    let hash = location.hash || '#/'
    let path = isHash
      ? hash
      : location.href.replace(location.origin, '').replace(hash, '')
    let query

    if (path.includes('?')) {
      ;[path, query] = path.split('?')
    }
    path = path.replace(PREFIX_REGEXP, '/')

    if (!$view || path === this.#route.path) {
      return
    }

    for (let [k, route] of this.#tables) {
      let args = path.match(route.regexp)
      if (args) {
        let params = Object.create(null)
        for (let i = 1; i < args.length; i++) {
          params[[route.vars[i - 1]]] = args[i]
        }
        let next = {
          path,
          name: route.name,
          params,
          query: query2object(query)
        }
        if (this.#beforeEach) {
          return this.#beforeEach(this.route, next, () => {
            this.#exec(next)
          })
        }

        return this.#exec(next)
      }
    }
    if (this.#tables.get('!')) {
      let route = this.#tables.get('!')
      $view.current = route.name
      this.#route = { path, name: route.name, params: {}, query: {} }
    }
  }

  #exec(route) {
    let $view = window.wkitd.get(__ROUTER_VIEW__)
    $view.current = route.name
    this.#route = route
    this.#rsync()
  }

  #rsync() {
    for (let [target, callback] of this.#targets) {
      callback.call(target, this.route)
    }
  }

  // 正式初始化路由监听, web components的特性, router-view未加载之前
  // 无法对路由进行操作, 所以需要另外在路由mounted时触发
  init() {
    this.#ready = true
    this.#hashchange()
  }

  /**
   * 用于同步路由到组件的
   */
  rsync(target, callback) {
    this.#targets.set(target, callback)
    // 路由已经初始化完成时, 还有新的同步请求则立刻执行
    if (this.#ready) {
      this.#rsync()
    }
  }

  beforeEach(callback = noop) {
    this.#beforeEach = callback
  }

  // 绑定路由事件
  addRoute(routes) {
    if (Array.isArray(routes)) {
      routes.forEach(it => {
        this.#add(it)
      })
    } else {
      this.#add(routes)
    }
    // 初始化后再添加路由, 手动执行一次回调
    if (this.#ready) {
      this.#hashchange()
    }
  }

  go(delta = 0) {
    history.go(delta)
  }

  back() {
    this.go(-1)
  }

  forward() {
    this.go(1)
  }

  //
  push(obj = { path: '', query: {} }, replace = false) {
    let path = '',
      query = ''
    if (typeof obj === 'string') {
      path = obj.trim()
    } else {
      query = object2query(obj.query || '')
      path = obj.path + (query ? `?${query}` : '')
    }

    // 空路径及相同路径, 不重复执行
    if (!path && path === location.hash.slice(1)) {
      return
    }

    if (this.type === MODE_HASH) {
      if (replace) {
        location.replace(path.replace(/^\//, '#/'))
      } else {
        location.hash = path
      }
    } else {
      if (replace) {
        window.history.replaceState({ path }, null, path + query)
      } else {
        window.history.pushState({ path }, null, path + query)
      }
      // pushState & replaceState 不会触发popstate事件
      // 要手动触发路由检测
      this.#check()
    }
  }

  //
  replace(obj = { path: '', query: {} }) {
    this.push(obj, true)
  }
}

export function createWebHashHistory() {
  return () => new Router()
}

export function createWebHistory() {
  return () => new Router(MODE_HISTORY)
}
